use crate::models::station_measurements::StationMeasurements;
use crate::utils::parse;
use memmap2::MmapOptions;
use std::sync::mpsc;
use std::time::Instant;
use std::{fs::File, thread};
use std::collections::HashMap;
use crate::utils::parse::hashstr;

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

pub fn run() {
    let now = Instant::now();
    const FILE_PATH: &str = "../../../measurements.txt";
    let file = File::open(FILE_PATH).expect("File measurements.txt not found");
    let mmap = unsafe { MmapOptions::new().map(&file).unwrap() };
    let file_length = mmap.len();
    let mut stations: HashMap<usize, (String, StationMeasurements)> =
        HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);
    let (tx, rx) = mpsc::channel();
    let cores = thread::available_parallelism().unwrap().into();
    let chunk_length = file_length / cores;
    let mut bounds = Vec::with_capacity(cores + 1);
    let mut start = 0;
    for _ in 0..cores {
        let end = (start + chunk_length).min(mmap.len());
        let next_new_line = match memchr::memchr(b'\n', &mmap[end..]) {
            Some(v) => v,
            None => {
                assert_eq!(end, mmap.len());
                0
            }
        };
        let end = end + next_new_line;
        bounds.push((start, end));
        start = end + 1;
    }
    thread::scope(|s| {
        for i in 0..cores {
            let tx = tx.clone();
            let (start, end) = *bounds.get(i).unwrap();
            let mmap_slice = &mmap[start..end];
            s.spawn(move || {
                let mut t_stations: HashMap<usize, (String, StationMeasurements)> =
                    HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);
                for line in mmap_slice.split(|&byte| { byte == b'\n' }) {
                    if line.len() == 0 {
                        break;
                    }
                    let (station, temp) = line.rsplit_once(|&byte| byte == b';').unwrap();
                    let hash = hashstr(station);
                    let station = unsafe { std::str::from_utf8_unchecked(station) };
                    let temp = parse::temp(temp);
                    let measurements_option = t_stations.get_mut(&hash);
                    if let Some((_, measurements)) = measurements_option {
                        measurements.update(temp);
                    } else {
                        let measurements = StationMeasurements {
                            min: temp,
                            max: temp,
                            count: 1,
                            sum: temp,
                        };
                        t_stations.insert(hash, (station.to_string(), measurements));
                    }
                }
                let _ = tx.send(t_stations);
            });
        }
        drop(tx);
        while let Ok(t_stations) = rx.recv() {
            for (hash, (station, measurements)) in t_stations.iter() {
                let joined_measurements_options = stations.get_mut(hash);
                if let Some((_, joined_measurements)) = joined_measurements_options {
                    joined_measurements.merge(measurements);
                } else {
                    stations.insert(*hash, (station.to_owned(), *measurements));
                }
            }
        }
        let mut stations: Vec<String> = stations
            .iter()
            .map(|(_, (station, measurements))| {
                let measurements = measurements.to_string();
                #[cfg(feature = "json")]
                {
                    format!("{{\"{station}\":\"{measurements}\"}}")
                }
                #[cfg(not(feature = "json"))]
                {
                    format!("{station}={measurements}")
                }
            })
            .collect();
        stations.sort();
        let stations = stations.join(",");
        #[cfg(feature = "json")]
        {
            println!("\n\n[{stations}]");
        }
        #[cfg(not(feature = "json"))]
        {
            println!("\n\n{{{stations}}}");
        }
        println!("\n\nTime={} ms", now.elapsed().as_millis());
    });
}
