use std::sync::mpsc;
use std::time::Instant;
use std::{fs::File, thread};
use std::ffi::CStr;
use memmap2::MmapOptions;
use rustc_hash::{FxBuildHasher, FxHashMap};
use crate::models::station_measurements::StationMeasurements;
use crate::utils::parse;

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

pub fn run() {
    print!("\x1b[J");
    const FILE_PATH: &str = "structured_measurements.txt";
    let now = Instant::now();
    let file = File::open(FILE_PATH).expect("File structured_measurements.txt not found");
    let mmap = unsafe { MmapOptions::new().map(&file).unwrap() };
    let file_length = mmap.len();
    let hasher = FxBuildHasher::default();
    let mut stations: FxHashMap<String, StationMeasurements> =
        FxHashMap::with_capacity_and_hasher(DEFAULT_HASHMAP_LENGTH, hasher);
    let (tx, rx) = mpsc::channel();
    let cores = thread::available_parallelism().unwrap().into();
    let chunk_length = file_length / cores;
    let mut bounds = Vec::with_capacity(cores + 1);
    let mut start = 0;
    for _ in 0..cores {
        let end = (start + chunk_length).min(mmap.len());
        let next_new_line = match memchr::memchr(b'\n', &mmap[end..]) {
            Some(v) => v,
            None => {
                assert_eq!(end, mmap.len());
                0
            }
        };
        let end = end + next_new_line;
        bounds.push((start, end));
        start = end + 1;
    }
    thread::scope(|s| {
        for i in 0..cores {
            let tx = tx.clone();
            let (start, end) = *bounds.get(i).unwrap();
            let mmap_slice = &mmap[start..end];
            s.spawn(move || {
                let mut t_stations: FxHashMap<String, StationMeasurements> =
                    FxHashMap::with_capacity_and_hasher(DEFAULT_HASHMAP_LENGTH, hasher);
                let lines = mmap_slice.chunks_exact(107);
                for (line_num, line) in lines.enumerate() {
                    if line_num % 100000 == 0 {
                        print!("\x1b[{i};0Hlines: {line_num}");
                    }
                    let (station, temp) = unsafe { line.split_at_unchecked(100) };
                    let station = {
                        if station[station.len() - 1] == 0u8 {
                            unsafe { std::str::from_utf8_unchecked(CStr::from_bytes_until_nul(station).unwrap().to_bytes()) }
                        } else {
                            unsafe { std::str::from_utf8_unchecked(station) }
                        }
                    }.to_owned();
                    let temp = parse::temp_new(&temp[1..6]);
                    let measurements_option = t_stations.get_mut(&station);
                    if let Some(measurements) = measurements_option {
                        measurements.update(temp);
                    } else {
                        let measurements = StationMeasurements {
                            min: temp,
                            max: temp,
                            count: 1,
                            sum: temp,
                        };
                        t_stations.insert(station, measurements);
                    }
                }
                let _ = tx.send(t_stations);
            });
        }
        drop(tx);
        while let Ok(t_stations) = rx.recv() {
            for (station, measurements) in t_stations.iter() {
                let joined_measurements_options = stations.get_mut(station);
                if let Some(joined_measurements) = joined_measurements_options {
                    joined_measurements.merge(measurements);
                } else {
                    stations.insert(station.to_owned(), *measurements);
                }
            }
        }
        let mut stations: Vec<String> = stations
            .iter()
            .map(|(station, measurements)| {
                let measurements = measurements.to_string();
                #[cfg(feature = "json")]
                {
                    format!("{{\"{station}\":\"{measurements}\"}}")
                }
                #[cfg(not(feature = "json"))]
                {
                    format!("{station}={measurements}")
                }
            })
            .collect();
        stations.sort();
        let stations = stations.join(",");
        #[cfg(feature = "json")]
        {
            println!("\n\n[{stations}]");
        }
        #[cfg(not(feature = "json"))]
        {
            println!("\n\n{{{stations}}}");
        }
        println!("\n\nTime={} ms", now.elapsed().as_millis());
    });
}
